from __future__ import absolute_import, division, print_function

from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import DeletableAPIResource
from stripe.api_resources.abstract import ListableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource


class SubscriptionItem(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "subscription_item"

    def usage_record_summaries(self, **params):
        return self.request(
            "get", self.instance_url() + "/usage_record_summaries", params
        )



from __future__ import absolute_import, division, print_function

import json

import pytest

import stripe


class TestListObject(object):
    @pytest.fixture
    def list_object(self):
        return stripe.ListObject.construct_from(
            {"object": "list", "url": "/my/path", "data": ["foo"]}, "mykey"
        )

    def assert_response(self, res):
        assert isinstance(res[0], stripe.Charge)
        assert res[0].foo == "bar"

    def test_for_loop(self, list_object):
        seen = []

        for item in list_object:
            seen.append(item)

        assert seen == ["foo"]

    def test_list(self, request_mock, list_object):
        request_mock.stub_request(
            "get", "/my/path", [{"object": "charge", "foo": "bar"}]
        )

        res = list_object.list(myparam="you")

        request_mock.assert_requested(
            "get", "/my/path", {"myparam": "you"}, None
        )
        self.assert_response(res)

    def test_create(self, request_mock, list_object):
        request_mock.stub_request(
            "post", "/my/path", [{"object": "charge", "foo": "bar"}]
        )

        res = list_object.create(myparam="eter")

        request_mock.assert_requested(
            "post", "/my/path", {"myparam": "eter"}, None
        )
        self.assert_response(res)

    def test_retrieve(self, request_mock, list_object):
        request_mock.stub_request(
            "get", "/my/path/myid", [{"object": "charge", "foo": "bar"}]
        )

        res = list_object.retrieve("myid", myparam="cow")

        request_mock.assert_requested(
            "get", "/my/path/myid", {"myparam": "cow"}, None
        )

        self.assert_response(res)

    def test_len(self, list_object):
        assert len(list_object) == 1

    def test_bool(self, list_object):
        assert list_object

        empty = stripe.ListObject.construct_from(
            {"object": "list", "url": "/my/path", "data": []}, "mykey"
        )
        assert bool(empty) is False

    def test_serialize_empty_list(self):
        empty = stripe.ListObject.construct_from(
            {"object": "list", "data": []}, "mykey"
        )
        serialized = str(empty)
        deserialized = stripe.ListObject.construct_from(
            json.loads(serialized), "mykey"
        )
        assert deserialized == empty

    def test_serialize_nested_empty_list(self):
        empty = stripe.ListObject.construct_from(
            {"object": "list", "data": []}, "mykey"
        )
        obj = stripe.stripe_object.StripeObject.construct_from(
            {"nested": empty}, "mykey"
        )
        serialized = str(obj)
        deserialized = stripe.stripe_object.StripeObject.construct_from(
            json.loads(serialized), "mykey"
        )
        assert deserialized.nested == empty


class TestAutoPaging:
    @staticmethod
    def pageable_model_response(ids, has_more):
        return {
            "object": "list",
            "url": "/v1/pageablemodels",
            "data": [{"id": id, "object": "pageablemodel"} for id in ids],
            "has_more": has_more,
        }

    def test_iter_one_page(self, request_mock):
        lo = stripe.ListObject.construct_from(
            self.pageable_model_response(["pm_123", "pm_124"], False), "mykey"
        )

        request_mock.assert_no_request()

        seen = [item["id"] for item in lo.auto_paging_iter()]

        assert seen == ["pm_123", "pm_124"]

    def test_iter_two_pages(self, request_mock):
        lo = stripe.ListObject.construct_from(
            self.pageable_model_response(["pm_123", "pm_124"], True), "mykey"
        )
        lo._retrieve_params = {"foo": "bar"}

        request_mock.stub_request(
            "get",
            "/v1/pageablemodels",
            self.pageable_model_response(["pm_125", "pm_126"], False),
        )

        seen = [item["id"] for item in lo.auto_paging_iter()]

        request_mock.assert_requested(
            "get",
            "/v1/pageablemodels",
            {"starting_after": "pm_124", "foo": "bar"},
            None,
        )

        assert seen == ["pm_123", "pm_124", "pm_125", "pm_126"]

    def test_class_method_two_pages(self, request_mock):
        request_mock.stub_request(
            "get",
            "/v1/charges",
            {
                "object": "list",
                "data": [{"id": "ch_001"}],
                "url": "/v1/charges",
                "has_more": False,
            },
        )

        seen = [
            item["id"]
            for item in stripe.Charge.auto_paging_iter(limit=25, foo="bar")
        ]

        request_mock.assert_requested(
            "get", "/v1/charges", {"limit": 25, "foo": "bar"}
        )
        assert seen == ["ch_001"]



from __future__ import absolute_import, division, print_function

import stripe


class TestDeletableAPIResource(object):
    class MyDeletable(stripe.api_resources.abstract.DeletableAPIResource):
        OBJECT_NAME = "mydeletable"

    def test_delete_class(self, request_mock):
        request_mock.stub_request(
            "delete",
            "/v1/mydeletables/mid",
            {"id": "mid", "deleted": True},
            rheaders={"request-id": "req_id"},
        )

        obj = self.MyDeletable.delete("mid")

        request_mock.assert_requested("delete", "/v1/mydeletables/mid", {})
        assert obj.deleted is True
        assert obj.id == "mid"

        assert obj.last_response is not None
        assert obj.last_response.request_id == "req_id"

    def test_delete_instance(self, request_mock):
        request_mock.stub_request(
            "delete",
            "/v1/mydeletables/mid",
            {"id": "mid", "deleted": True},
            rheaders={"request-id": "req_id"},
        )

        obj = self.MyDeletable.construct_from({"id": "mid"}, "mykey")

        assert obj is obj.delete()
        request_mock.assert_requested("delete", "/v1/mydeletables/mid", {})
        assert obj.deleted is True
        assert obj.id == "mid"

        assert obj.last_response is not None
        assert obj.last_response.request_id == "req_id"



from __future__ import absolute_import, division, print_function

from stripe import error
from stripe import util
from stripe.api_resources.abstract import DeletableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource
from stripe.api_resources.account import Account
from stripe.api_resources.customer import Customer
from stripe.api_resources.recipient import Recipient
from stripe.six.moves.urllib.parse import quote_plus


class Card(DeletableAPIResource, UpdateableAPIResource):
    OBJECT_NAME = "card"

    def instance_url(self):
        token = util.utf8(self.id)
        extn = quote_plus(token)
        if hasattr(self, "customer"):
            customer = util.utf8(self.customer)

            base = Customer.class_url()
            owner_extn = quote_plus(customer)
            class_base = "sources"

        elif hasattr(self, "recipient"):
            recipient = util.utf8(self.recipient)

            base = Recipient.class_url()
            owner_extn = quote_plus(recipient)
            class_base = "cards"

        elif hasattr(self, "account"):
            account = util.utf8(self.account)

            base = Account.class_url()
            owner_extn = quote_plus(account)
            class_base = "external_accounts"

        else:
            raise error.InvalidRequestError(
                "Could not determine whether card_id %s is "
                "attached to a customer, recipient, or "
                "account." % token,
                "id",
            )

        return "%s/%s/%s/%s" % (base, owner_extn, class_base, extn)

    @classmethod
    def modify(cls, sid, **params):
        raise NotImplementedError(
            "Can't modify a card without a customer, recipient or account "
            "ID. Call save on customer.sources.retrieve('card_id'), "
            "recipient.cards.retrieve('card_id'), or "
            "account.external_accounts.retrieve('card_id') instead."
        )

    @classmethod
    def retrieve(
        cls,
        id,
        api_key=None,
        stripe_version=None,
        stripe_account=None,
        **params
    ):
        raise NotImplementedError(
            "Can't retrieve a card without a customer, recipient or account "
            "ID. Use customer.sources.retrieve('card_id'), "
            "recipient.cards.retrieve('card_id'), or "
            "account.external_accounts.retrieve('card_id') instead."
        )



from __future__ import absolute_import, division, print_function

from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import ListableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource


class Cardholder(
    CreateableAPIResource, ListableAPIResource, UpdateableAPIResource
):
    OBJECT_NAME = "issuing.cardholder"



from __future__ import absolute_import, division, print_function

import stripe


TEST_RESOURCE_ID = "tr_123"
TEST_REVERSAL_ID = "trr_123"


class TestTransfer(object):
    def test_is_listable(self, request_mock):
        resources = stripe.Transfer.list()
        request_mock.assert_requested("get", "/v1/transfers")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], stripe.Transfer)

    def test_is_retrievable(self, request_mock):
        resource = stripe.Transfer.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v1/transfers/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, stripe.Transfer)

    def test_is_creatable(self, request_mock):
        resource = stripe.Transfer.create(
            amount=100, currency="usd", destination="acct_123"
        )
        request_mock.assert_requested("post", "/v1/transfers")
        assert isinstance(resource, stripe.Transfer)

    def test_is_saveable(self, request_mock):
        resource = stripe.Transfer.retrieve(TEST_RESOURCE_ID)
        resource.metadata["key"] = "value"
        resource.save()
        request_mock.assert_requested(
            "post", "/v1/transfers/%s" % TEST_RESOURCE_ID
        )

    def test_is_modifiable(self, request_mock):
        resource = stripe.Transfer.modify(
            TEST_RESOURCE_ID, metadata={"key": "value"}
        )
        request_mock.assert_requested(
            "post", "/v1/transfers/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, stripe.Transfer)

    def test_can_cancel(self, request_mock):
        # stripe-mock does not handle this anymore as it was on an old
        # API version so we stub instead.
        request_mock.stub_request(
            "post",
            "/v1/transfers/%s/cancel" % TEST_RESOURCE_ID,
            {
                "id": "%s" % TEST_RESOURCE_ID,
                "object": "transfer",
                "status": "canceled",
            },
        )
        transfer = stripe.Transfer.construct_from(
            {"id": "%s" % TEST_RESOURCE_ID, "object": "transfer"},
            stripe.api_key,
        )
        transfer_canceled = transfer.cancel()
        request_mock.assert_requested(
            "post", "/v1/transfers/%s/cancel" % TEST_RESOURCE_ID
        )
        assert isinstance(transfer_canceled, stripe.Transfer)

    def test_can_cancel_classmethod(self, request_mock):
        # stripe-mock does not handle this anymore as it was on an old
        # API version so we stub instead.
        request_mock.stub_request(
            "post",
            "/v1/transfers/%s/cancel" % TEST_RESOURCE_ID,
            {"id": "%s" % TEST_RESOURCE_ID, "object": "transfer"},
        )
        transfer = stripe.Transfer.cancel(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "post", "/v1/transfers/%s/cancel" % TEST_RESOURCE_ID
        )
        assert isinstance(transfer, stripe.Transfer)


class TestTransferReversals:
    def test_is_listable(self, request_mock):
        resources = stripe.Transfer.list_reversals(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v1/transfers/%s/reversals" % TEST_RESOURCE_ID
        )
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], stripe.Reversal)

    def test_is_retrievable(self, request_mock):
        resource = stripe.Transfer.retrieve_reversal(
            TEST_RESOURCE_ID, TEST_REVERSAL_ID
        )
        request_mock.assert_requested(
            "get",
            "/v1/transfers/%s/reversals/%s"
            % (TEST_RESOURCE_ID, TEST_REVERSAL_ID),
        )
        assert isinstance(resource, stripe.Reversal)

    def test_is_creatable(self, request_mock):
        resource = stripe.Transfer.create_reversal(
            TEST_RESOURCE_ID, amount=100
        )
        request_mock.assert_requested(
            "post", "/v1/transfers/%s/reversals" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, stripe.Reversal)

    def test_is_modifiable(self, request_mock):
        resource = stripe.Transfer.modify_reversal(
            TEST_RESOURCE_ID, TEST_REVERSAL_ID, metadata={"foo": "bar"}
        )
        request_mock.assert_requested(
            "post",
            "/v1/transfers/%s/reversals/%s"
            % (TEST_RESOURCE_ID, TEST_REVERSAL_ID),
        )
        assert isinstance(resource, stripe.Reversal)



from __future__ import absolute_import, division, print_function

import pytest

import stripe


TEST_RESOURCE_ID = "sub_sched_rev_123"


class TestSubscriptionScheduleRevision(object):
    def construct_resource(self):
        revision_dict = {
            "id": TEST_RESOURCE_ID,
            "object": "subscription_schedule_revision",
            "schedule": "sub_sched_123",
        }
        return stripe.SubscriptionScheduleRevision.construct_from(
            revision_dict, stripe.api_key
        )

    def test_has_instance_url(self, request_mock):
        resource = self.construct_resource()
        assert (
            resource.instance_url()
            == "/v1/subscription_schedules/sub_sched_123/revisions/%s"
            % TEST_RESOURCE_ID
        )

    def test_is_not_retrievable(self, request_mock):
        with pytest.raises(NotImplementedError):
            stripe.SubscriptionScheduleRevision.retrieve(TEST_RESOURCE_ID)



from __future__ import absolute_import, division, print_function

from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import DeletableAPIResource
from stripe.api_resources.abstract import ListableAPIResource


class ApplePayDomain(
    CreateableAPIResource, DeletableAPIResource, ListableAPIResource
):
    OBJECT_NAME = "apple_pay_domain"

    @classmethod
    def class_url(cls):
        return "/v1/apple_pay/domains"



from __future__ import absolute_import, division, print_function

from stripe.api_resources.abstract import CreateableAPIResource


class ThreeDSecure(CreateableAPIResource):
    OBJECT_NAME = "three_d_secure"

    @classmethod
    def class_url(cls):
        return "/v1/3d_secure"



from __future__ import absolute_import, division, print_function

from stripe import util


class VerifyMixin(object):
    def verify(self, idempotency_key=None, **params):
        url = self.instance_url() + "/verify"
        headers = util.populate_headers(idempotency_key)
        self.refresh_from(self.request("post", url, params, headers))
        return self



from __future__ import absolute_import, division, print_function

import stripe


TEST_RESOURCE_ID = "txn_123"


class TestBalanceTransaction(object):
    def test_is_listable(self, request_mock):
        resources = stripe.BalanceTransaction.list()
        request_mock.assert_requested("get", "/v1/balance/history")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], stripe.BalanceTransaction)



from __future__ import absolute_import, division, print_function

from six.moves.urllib.parse import parse_qs, urlparse

import stripe


class TestOAuth(object):
    def test_authorize_url(self, request_mock):
        url = stripe.OAuth.authorize_url(
            scope="read_write",
            state="csrf_token",
            stripe_user={
                "email": "test@example.com",
                "url": "https://example.com/profile/test",
                "country": "US",
            },
        )

        o = urlparse(url)
        params = parse_qs(o.query)

        assert o.scheme == "https"
        assert o.netloc == "connect.stripe.com"
        assert o.path == "/oauth/authorize"

        assert params["client_id"] == ["ca_123"]
        assert params["scope"] == ["read_write"]
        assert params["stripe_user[email]"] == ["test@example.com"]
        assert params["stripe_user[url]"] == [
            "https://example.com/profile/test"
        ]
        assert params["stripe_user[country]"] == ["US"]

    def test_token(self, request_mock):
        request_mock.stub_request(
            "post",
            "/oauth/token",
            {
                "access_token": "sk_access_token",
                "scope": "read_only",
                "livemode": "false",
                "token_type": "bearer",
                "refresh_token": "sk_refresh_token",
                "stripe_user_id": "acct_test",
                "stripe_publishable_key": "pk_test",
            },
        )

        resp = stripe.OAuth.token(
            grant_type="authorization_code",
            code="this_is_an_authorization_code",
        )
        request_mock.assert_requested(
            "post",
            "/oauth/token",
            {
                "grant_type": "authorization_code",
                "code": "this_is_an_authorization_code",
            },
        )
        assert resp["access_token"] == "sk_access_token"

    def test_deauthorize(self, request_mock):
        request_mock.stub_request(
            "post",
            "/oauth/deauthorize",
            {"stripe_user_id": "acct_test_deauth"},
        )

        resp = stripe.OAuth.deauthorize(stripe_user_id="acct_test_deauth")
        request_mock.assert_requested(
            "post",
            "/oauth/deauthorize",
            {"client_id": "ca_123", "stripe_user_id": "acct_test_deauth"},
        )
        assert resp["stripe_user_id"] == "acct_test_deauth"



from __future__ import absolute_import, division, print_function

import json

from stripe.stripe_response import StripeResponse


class TestStripeResponse(object):
    def test_idempotency_key(self):
        response, headers, _, _ = self.mock_stripe_response()
        assert response.idempotency_key == headers["idempotency-key"]

    def test_request_id(self):
        response, headers, _, _ = self.mock_stripe_response()
        assert response.request_id == headers["request-id"]

    def test_code(self):
        response, _, _, code = self.mock_stripe_response()
        assert response.code == code

    def test_headers(self):
        response, headers, _, _ = self.mock_stripe_response()
        assert response.headers == headers

    def test_body(self):
        response, _, body, _ = self.mock_stripe_response()
        assert response.body == body

    def test_data(self):
        response, _, body, _ = self.mock_stripe_response()
        assert response.data == json.loads(body)

    @staticmethod
    def mock_stripe_response():
        code = 200
        headers = TestStripeResponse.mock_headers()
        body = TestStripeResponse.mock_body()
        response = StripeResponse(body, code, headers)
        return response, headers, body, code

    @staticmethod
    def mock_headers():
        return {"idempotency-key": "123456", "request-id": "req_123456"}

    @staticmethod
    def mock_body():
        return '{ "id": "ch_12345", "object": "charge", "amount": 1 }'



from __future__ import absolute_import, division, print_function

import stripe


TEST_RESOURCE_ID = "idp_123"


class TestDispute(object):
    def test_is_creatable(self, request_mock):
        resource = stripe.issuing.Dispute.create(
            reason="fraudulent", disputed_transaction="ipi_123"
        )
        request_mock.assert_requested("post", "/v1/issuing/disputes")
        assert isinstance(resource, stripe.issuing.Dispute)

    def test_is_listable(self, request_mock):
        resources = stripe.issuing.Dispute.list()
        request_mock.assert_requested("get", "/v1/issuing/disputes")
        assert isinstance(resources.data, list)
        assert isinstance(resources.data[0], stripe.issuing.Dispute)

    def test_is_modifiable(self, request_mock):
        resource = stripe.issuing.Dispute.modify(
            TEST_RESOURCE_ID, metadata={"key": "value"}
        )
        request_mock.assert_requested(
            "post", "/v1/issuing/disputes/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, stripe.issuing.Dispute)

    def test_is_retrievable(self, request_mock):
        resource = stripe.issuing.Dispute.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v1/issuing/disputes/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, stripe.issuing.Dispute)

    def test_is_saveable(self, request_mock):
        resource = stripe.issuing.Dispute.retrieve(TEST_RESOURCE_ID)
        resource.metadata["key"] = "value"
        dispute = resource.save()
        request_mock.assert_requested(
            "post", "/v1/issuing/disputes/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, stripe.issuing.Dispute)
        assert resource is dispute



from __future__ import absolute_import, division, print_function

import pytest

import stripe


TEST_RESOURCE_ID = "src_123"


class TestSource(object):
    def test_is_retrievable(self, request_mock):
        resource = stripe.Source.retrieve(TEST_RESOURCE_ID)
        request_mock.assert_requested(
            "get", "/v1/sources/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, stripe.Source)

    def test_is_creatable(self, request_mock):
        resource = stripe.Source.create(type="card", token="tok_123")
        request_mock.assert_requested("post", "/v1/sources")
        assert isinstance(resource, stripe.Source)

    def test_is_saveable(self, request_mock):
        resource = stripe.Source.retrieve(TEST_RESOURCE_ID)
        resource.metadata["key"] = "value"
        resource.save()
        request_mock.assert_requested(
            "post", "/v1/sources/%s" % TEST_RESOURCE_ID
        )

    def test_is_modifiable(self, request_mock):
        resource = stripe.Source.modify(
            TEST_RESOURCE_ID, metadata={"key": "value"}
        )
        request_mock.assert_requested(
            "post", "/v1/sources/%s" % TEST_RESOURCE_ID
        )
        assert isinstance(resource, stripe.Source)

    def test_is_detachable_when_attached(self, request_mock):
        resource = stripe.Source.construct_from(
            {
                "id": TEST_RESOURCE_ID,
                "object": "source",
                "customer": "cus_123",
            },
            stripe.api_key,
        )
        source = resource.detach()
        assert source is resource
        request_mock.assert_requested(
            "delete", "/v1/customers/cus_123/sources/%s" % TEST_RESOURCE_ID
        )

    def test_is_not_detachable_when_unattached(self):
        resource = stripe.Source.retrieve(TEST_RESOURCE_ID)
        with pytest.raises(stripe.error.InvalidRequestError):
            resource.detach()

    def test_is_verifiable(self, request_mock):
        resource = stripe.Source.retrieve(TEST_RESOURCE_ID)
        source = resource.verify(values=[1, 2])
        assert source is resource
        request_mock.assert_requested(
            "post",
            "/v1/sources/%s/verify" % TEST_RESOURCE_ID,
            {"values": [1, 2]},
        )



from __future__ import absolute_import, division, print_function

import stripe


class TestCreateableAPIResource(object):
    class MyCreatable(stripe.api_resources.abstract.CreateableAPIResource):
        OBJECT_NAME = "mycreatable"

    def test_create(self, request_mock):
        request_mock.stub_request(
            "post",
            "/v1/mycreatables",
            {"object": "charge", "foo": "bar"},
            rheaders={"request-id": "req_id"},
        )

        res = self.MyCreatable.create()

        request_mock.assert_requested("post", "/v1/mycreatables", {})
        assert isinstance(res, stripe.Charge)
        assert res.foo == "bar"

        assert res.last_response is not None
        assert res.last_response.request_id == "req_id"

    def test_idempotent_create(self, request_mock):
        request_mock.stub_request(
            "post",
            "/v1/mycreatables",
            {"object": "charge", "foo": "bar"},
            rheaders={"idempotency-key": "foo"},
        )

        res = self.MyCreatable.create(idempotency_key="foo")

        request_mock.assert_requested(
            "post", "/v1/mycreatables", {}, {"Idempotency-Key": "foo"}
        )
        assert isinstance(res, stripe.Charge)
        assert res.foo == "bar"



import os
import sys
from codecs import open
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = "-n auto"

    def run_tests(self):
        import shlex

        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


here = os.path.abspath(os.path.dirname(__file__))

os.chdir(here)

with open(
    os.path.join(here, "LONG_DESCRIPTION.rst"), "r", encoding="utf-8"
) as fp:
    long_description = fp.read()

version_contents = {}
with open(os.path.join(here, "stripe", "version.py"), encoding="utf-8") as f:
    exec(f.read(), version_contents)

setup(
    name="stripe",
    version=version_contents["VERSION"],
    description="Python bindings for the Stripe API",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Stripe",
    author_email="support@stripe.com",
    url="https://github.com/stripe/stripe-python",
    license="MIT",
    keywords="stripe api payments",
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={"stripe": ["data/ca-certificates.crt"]},
    zip_safe=False,
    install_requires=[
        'requests >= 2.20; python_version >= "3.0"',
        'requests[security] >= 2.20; python_version < "3.0"',
    ],
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    tests_require=[
        "pytest >= 4.6.2, < 4.7",
        "pytest-mock >= 1.10.4",
        "pytest-xdist >= 1.28.0",
        "pytest-cov >= 2.7.1",
        # coverage 5.0 pre-releases don't work, and setuptools doesn't ignore
        # pre-releases (cf. https://github.com/pypa/setuptools/issues/855)
        "coverage >= 4.5.3, < 5",
    ],
    cmdclass={"test": PyTest},
    project_urls={
        "Bug Tracker": "https://github.com/stripe/stripe-python/issues",
        "Documentation": "https://stripe.com/docs/api/python",
        "Source Code": "https://github.com/stripe/stripe-python",
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from __future__ import absolute_import, division, print_function

from stripe import util
from stripe.api_resources.abstract import CreateableAPIResource
from stripe.api_resources.abstract import ListableAPIResource
from stripe.api_resources.abstract import UpdateableAPIResource
from stripe.api_resources.abstract import custom_method


@custom_method("void_credit_note", http_verb="post", http_path="void")
class CreditNote(
    CreateableAPIResource, ListableAPIResource, UpdateableAPIResource
):
    OBJECT_NAME = "credit_note"

    def void_credit_note(self, idempotency_key=None, **params):
        url = self.instance_url() + "/void"
        headers = util.populate_headers(idempotency_key)
        self.refresh_from(self.request("post", url, params, headers))
        return self



from __future__ import absolute_import, division, print_function

import calendar
import datetime
import json
import platform
import time
import uuid
import warnings

import stripe
from stripe import error, oauth_error, http_client, version, util, six
from stripe.multipart_data_generator import MultipartDataGenerator
from stripe.six.moves.urllib.parse import urlencode, urlsplit, urlunsplit
from stripe.stripe_response import StripeResponse


def _encode_datetime(dttime):
    if dttime.tzinfo and dttime.tzinfo.utcoffset(dttime) is not None:
        utc_timestamp = calendar.timegm(dttime.utctimetuple())
    else:
        utc_timestamp = time.mktime(dttime.timetuple())

    return int(utc_timestamp)


def _encode_nested_dict(key, data, fmt="%s[%s]"):
    d = {}
    for subkey, subvalue in six.iteritems(data):
        d[fmt % (key, subkey)] = subvalue
    return d


def _api_encode(data):
    for key, value in six.iteritems(data):
        key = util.utf8(key)
        if value is None:
            continue
        elif hasattr(value, "stripe_id"):
            yield (key, value.stripe_id)
        elif isinstance(value, list) or isinstance(value, tuple):
            for i, sv in enumerate(value):
                if isinstance(sv, dict):
                    subdict = _encode_nested_dict("%s[%d]" % (key, i), sv)
                    for k, v in _api_encode(subdict):
                        yield (k, v)
                else:
                    yield ("%s[%d]" % (key, i), util.utf8(sv))
        elif isinstance(value, dict):
            subdict = _encode_nested_dict(key, value)
            for subkey, subvalue in _api_encode(subdict):
                yield (subkey, subvalue)
        elif isinstance(value, datetime.datetime):
            yield (key, _encode_datetime(value))
        else:
            yield (key, util.utf8(value))


def _build_api_url(url, query):
    scheme, netloc, path, base_query, fragment = urlsplit(url)

    if base_query:
        query = "%s&%s" % (base_query, query)

    return urlunsplit((scheme, netloc, path, query, fragment))


class APIRequestor(object):
    def __init__(
        self,
        key=None,
        client=None,
        api_base=None,
        api_version=None,
        account=None,
    ):
        self.api_base = api_base or stripe.api_base
        self.api_key = key
        self.api_version = api_version or stripe.api_version
        self.stripe_account = account

        self._default_proxy = None

        from stripe import verify_ssl_certs as verify
        from stripe import proxy

        if client:
            self._client = client
        elif stripe.default_http_client:
            self._client = stripe.default_http_client
            if proxy != self._default_proxy:
                warnings.warn(
                    "stripe.proxy was updated after sending a "
                    "request - this is a no-op. To use a different proxy, "
                    "set stripe.default_http_client to a new client "
                    "configured with the proxy."
                )
        else:
            # If the stripe.default_http_client has not been set by the user
            # yet, we'll set it here. This way, we aren't creating a new
            # HttpClient for every request.
            stripe.default_http_client = http_client.new_default_http_client(
                verify_ssl_certs=verify, proxy=proxy
            )
            self._client = stripe.default_http_client
            self._default_proxy = proxy

    @classmethod
    def format_app_info(cls, info):
        str = info["name"]
        if info["version"]:
            str += "/%s" % (info["version"],)
        if info["url"]:
            str += " (%s)" % (info["url"],)
        return str

    def request(self, method, url, params=None, headers=None):
        rbody, rcode, rheaders, my_api_key = self.request_raw(
            method.lower(), url, params, headers
        )
        resp = self.interpret_response(rbody, rcode, rheaders)
        return resp, my_api_key

    def handle_error_response(self, rbody, rcode, resp, rheaders):
        try:
            error_data = resp["error"]
        except (KeyError, TypeError):
            raise error.APIError(
                "Invalid response object from API: %r (HTTP response code "
                "was %d)" % (rbody, rcode),
                rbody,
                rcode,
                resp,
            )

        err = None

        # OAuth errors are a JSON object where `error` is a string. In
        # contrast, in API errors, `error` is a hash with sub-keys. We use
        # this property to distinguish between OAuth and API errors.
        if isinstance(error_data, six.string_types):
            err = self.specific_oauth_error(
                rbody, rcode, resp, rheaders, error_data
            )

        if err is None:
            err = self.specific_api_error(
                rbody, rcode, resp, rheaders, error_data
            )

        raise err

    def specific_api_error(self, rbody, rcode, resp, rheaders, error_data):
        util.log_info(
            "Stripe API error received",
            error_code=error_data.get("code"),
            error_type=error_data.get("type"),
            error_message=error_data.get("message"),
            error_param=error_data.get("param"),
        )

        # Rate limits were previously coded as 400's with code 'rate_limit'
        if rcode == 429 or (
            rcode == 400 and error_data.get("code") == "rate_limit"
        ):
            return error.RateLimitError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode in [400, 404]:
            if error_data.get("type") == "idempotency_error":
                return error.IdempotencyError(
                    error_data.get("message"), rbody, rcode, resp, rheaders
                )
            else:
                return error.InvalidRequestError(
                    error_data.get("message"),
                    error_data.get("param"),
                    error_data.get("code"),
                    rbody,
                    rcode,
                    resp,
                    rheaders,
                )
        elif rcode == 401:
            return error.AuthenticationError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode == 402:
            return error.CardError(
                error_data.get("message"),
                error_data.get("param"),
                error_data.get("code"),
                rbody,
                rcode,
                resp,
                rheaders,
            )
        elif rcode == 403:
            return error.PermissionError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        else:
            return error.APIError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )

    def specific_oauth_error(self, rbody, rcode, resp, rheaders, error_code):
        description = resp.get("error_description", error_code)

        util.log_info(
            "Stripe OAuth error received",
            error_code=error_code,
            error_description=description,
        )

        args = [error_code, description, rbody, rcode, resp, rheaders]

        if error_code == "invalid_client":
            return oauth_error.InvalidClientError(*args)
        elif error_code == "invalid_grant":
            return oauth_error.InvalidGrantError(*args)
        elif error_code == "invalid_request":
            return oauth_error.InvalidRequestError(*args)
        elif error_code == "invalid_scope":
            return oauth_error.InvalidScopeError(*args)
        elif error_code == "unsupported_grant_type":
            return oauth_error.UnsupportedGrantTypError(*args)
        elif error_code == "unsupported_response_type":
            return oauth_error.UnsupportedResponseTypError(*args)

        return None

    def request_headers(self, api_key, method):
        user_agent = "Stripe/v1 PythonBindings/%s" % (version.VERSION,)
        if stripe.app_info:
            user_agent += " " + self.format_app_info(stripe.app_info)

        ua = {
            "bindings_version": version.VERSION,
            "lang": "python",
            "publisher": "stripe",
            "httplib": self._client.name,
        }
        for attr, func in [
            ["lang_version", platform.python_version],
            ["platform", platform.platform],
            ["uname", lambda: " ".join(platform.uname())],
        ]:
            try:
                val = func()
            except Exception as e:
                val = "!! %s" % (e,)
            ua[attr] = val
        if stripe.app_info:
            ua["application"] = stripe.app_info

        headers = {
            "X-Stripe-Client-User-Agent": json.dumps(ua),
            "User-Agent": user_agent,
            "Authorization": "Bearer %s" % (api_key,),
        }

        if self.stripe_account:
            headers["Stripe-Account"] = self.stripe_account

        if method == "post":
            headers["Content-Type"] = "application/x-www-form-urlencoded"
            headers.setdefault("Idempotency-Key", str(uuid.uuid4()))

        if self.api_version is not None:
            headers["Stripe-Version"] = self.api_version

        return headers

    def request_raw(self, method, url, params=None, supplied_headers=None):
        """
        Mechanism for issuing an API call
        """

        if self.api_key:
            my_api_key = self.api_key
        else:
            from stripe import api_key

            my_api_key = api_key

        if my_api_key is None:
            raise error.AuthenticationError(
                "No API key provided. (HINT: set your API key using "
                '"stripe.api_key = <API-KEY>"). You can generate API keys '
                "from the Stripe web interface.  See https://stripe.com/api "
                "for details, or email support@stripe.com if you have any "
                "questions."
            )

        abs_url = "%s%s" % (self.api_base, url)

        encoded_params = urlencode(list(_api_encode(params or {})))

        # Don't use strict form encoding by changing the square bracket control
        # characters back to their literals. This is fine by the server, and
        # makes these parameter strings easier to read.
        encoded_params = encoded_params.replace("%5B", "[").replace("%5D", "]")

        if method == "get" or method == "delete":
            if params:
                abs_url = _build_api_url(abs_url, encoded_params)
            post_data = None
        elif method == "post":
            if (
                supplied_headers is not None
                and supplied_headers.get("Content-Type")
                == "multipart/form-data"
            ):
                generator = MultipartDataGenerator()
                generator.add_params(params or {})
                post_data = generator.get_post_data()
                supplied_headers[
                    "Content-Type"
                ] = "multipart/form-data; boundary=%s" % (generator.boundary,)
            else:
                post_data = encoded_params
        else:
            raise error.APIConnectionError(
                "Unrecognized HTTP method %r.  This may indicate a bug in the "
                "Stripe bindings.  Please contact support@stripe.com for "
                "assistance." % (method,)
            )

        headers = self.request_headers(my_api_key, method)
        if supplied_headers is not None:
            for key, value in six.iteritems(supplied_headers):
                headers[key] = value

        util.log_info("Request to Stripe api", method=method, path=abs_url)
        util.log_debug(
            "Post details",
            post_data=encoded_params,
            api_version=self.api_version,
        )

        rbody, rcode, rheaders = self._client.request_with_retries(
            method, abs_url, headers, post_data
        )

        util.log_info("Stripe API response", path=abs_url, response_code=rcode)
        util.log_debug("API response body", body=rbody)

        if "Request-Id" in rheaders:
            request_id = rheaders["Request-Id"]
            util.log_debug(
                "Dashboard link for request",
                link=util.dashboard_link(request_id),
            )

        return rbody, rcode, rheaders, my_api_key

    def interpret_response(self, rbody, rcode, rheaders):
        try:
            if hasattr(rbody, "decode"):
                rbody = rbody.decode("utf-8")
            resp = StripeResponse(rbody, rcode, rheaders)
        except Exception:
            raise error.APIError(
                "Invalid response body from API: %s "
                "(HTTP response code was %d)" % (rbody, rcode),
                rbody,
                rcode,
                rheaders,
            )
        if not 200 <= rcode < 300:
            self.handle_error_response(rbody, rcode, resp.data, rheaders)

        return resp



from __future__ import absolute_import, division, print_function

from stripe import util
from stripe.api_resources.abstract.api_resource import APIResource
from stripe.api_resources.subscription_schedule import SubscriptionSchedule
from stripe.six.moves.urllib.parse import quote_plus


class SubscriptionScheduleRevision(APIResource):
    OBJECT_NAME = "subscription_schedule_revision"

    def instance_url(self):
        token = util.utf8(self.id)
        schedule = util.utf8(self.schedule)
        base = SubscriptionSchedule.class_url()
        schedule_extn = quote_plus(schedule)
        extn = quote_plus(token)
        return "%s/%s/revisions/%s" % (base, schedule_extn, extn)

    @classmethod
    def retrieve(cls, id, api_key=None, **params):
        raise NotImplementedError(
            "Can't retrieve a subscription schedule revision without a schedule "
            "ID. Use schedule.revisions.retrieve('revision_id')"
        )
